# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from typing import Any, Dict, List, Optional

import boto3
import botocore
from fbpcp.decorator.error_handler import error_handler
from fbpcp.error.pcp import PcpError
from fbpcp.gateway.aws import AWSGateway


class TagsGateway(AWSGateway):
    def __init__(
        self,
        region: str,
        access_key_id: Optional[str] = None,
        access_key_data: Optional[str] = None,
        config: Optional[Dict[str, Any]] = None,
    ) -> None:
        super().__init__(region, access_key_id, access_key_data, config)

        self.client: botocore.client.BaseClient = boto3.client(
            "resourcegroupstaggingapi", region_name=self.region, **self.config
        )

    @error_handler
    def tag_resources(self, resource_arn_list: List[str], tags: Dict[str, str]) -> None:
        response = self.client.tag_resources(
            ResourceARNList=resource_arn_list, Tags=tags
        )
        failures = response["FailedResourcesMap"]
        if failures:
            raise PcpError(f"Tagging unsuccessful. Failed Resources: {failures}")

    @error_handler
    def untag_resources(
        self, resource_arn_list: List[str], tag_keys: List[str]
    ) -> None:
        response = self.client.untag_resources(
            ResourceARNList=resource_arn_list, TagKeys=tag_keys
        )
        failures = response["FailedResourcesMap"]
        if failures:
            raise PcpError(f"Untagging unsuccessful. Failed Resources: {failures}")

    @error_handler
    def get_resources_for_tag(self, tag_name: str, tag_value: str) -> List[str]:
        """Returns the resource ARNs for the resources matching the given tag name and value (upto 100)"""

        tag_filter = {"Key": tag_name, "Values": [tag_value]}
        response = self.client.get_resources(TagFilters=[tag_filter])
        resource_arn_list = [
            resource["ResourceARN"] for resource in response["ResourceTagMappingList"]
        ]
        return resource_arn_list



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import abc
from typing import Optional


class CoreDumpHandler(abc.ABC):
    @abc.abstractmethod
    def locate_core_dump_file(self) -> Optional[str]:
        """Find the absolute path of the core dump file
        return None if not found.
        """
        pass

    @abc.abstractmethod
    def upload_core_dump_file(self, core_dump_file_path: str, upload_dest: str) -> None:
        """Upload the core dump file to the specified location.
        Keyword arguments:
        core_dump_file_path: absolute path of the generated core dump file
        upload_dest: path to the cloud storage. e.g AWS S3 bucket
        """
        pass



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from hashlib import new
from typing import Dict, List

from fbpcp.service.storage import PathType, StorageService
from onedocker.entity.checksum_type import ChecksumType


class LocalChecksumGenerator:
    def _get_checksum(
        self,
        contents: bytes,
        checksum_algorithm: str,
    ) -> Dict[str, str]:
        hash_function = new(checksum_algorithm.lower(), contents)
        return {checksum_algorithm: hash_function.hexdigest()}

    def _read_local_file(self, path: str) -> bytes:
        if StorageService.path_type(path) != PathType.Local:
            raise ValueError("Only supports local paths")

        try:
            with open(path, "rb") as file:
                contents = file.read()
        except FileNotFoundError:
            raise ValueError("Please ensure that file exists at path specified")
        return contents

    def generate_checksums(
        self,
        binary_path: str,
        checksum_algorithms: List[ChecksumType],
    ) -> Dict[str, str]:
        """
        Usage:
            generate_checksums("/usr/bin/ls")
        """
        if len(checksum_algorithms) == 0:
            raise ValueError("No hashing function(s) have been provided")
        encoded_contents = self._read_local_file(binary_path)
        checksums = {}
        for checksum_algorithm in checksum_algorithms:
            checksums.update(
                self._get_checksum(
                    encoded_contents,
                    checksum_algorithm.name,
                )
            )
        return checksums



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict
import logging
import os

from cryptography import x509
from cryptography.x509.oid import NameOID
from fbpcp.entity.certificate_request import CertificateRequest
from onedocker.gateway.cryptography import CryptographyGateway
from onedocker.mapper.cryptography import (
    map_certificaterequest_to_x509name,
    map_certificaterequest_to_x509subjectalternativename,
)
from onedocker.service.certificate import CertificateService

DEFAULT_CERT_FOLDER = "certificate"
DEFAULT_PRIVATE_KEY_NAME = "private_key.pem"
DEFAULT_CERTIFICATE_NAME = "certificate.pem"
DEFAULT_DAYS_VALID = 5

ISSUER_COUNTRY_NAME = "US"
ISSUER_ORGANIZATION_NAME = "OneDocker"


def convert_str_to_bytes(s: str) -> bytes:
    return s.encode("utf-8")


class SelfSignedCertificateService(CertificateService):
    def __init__(self, cert_request: CertificateRequest, exe_path: str) -> None:
        self.logger: logging.Logger = logging.getLogger(__name__)
        self._cert_request = cert_request
        self._cert_path: str = os.path.join(
            exe_path, self._cert_request.cert_folder or DEFAULT_CERT_FOLDER
        )
        self._private_key_file: str = self._get_file_path(
            self._cert_request.private_key_name or DEFAULT_PRIVATE_KEY_NAME,
        )
        self._certificate_file: str = self._get_file_path(
            self._cert_request.certificate_name or DEFAULT_CERTIFICATE_NAME
        )
        self._crypt_gateway: CryptographyGateway = CryptographyGateway()

    def _get_file_path(self, filename: str) -> str:
        return os.path.join(self._cert_path, filename)

    def _write_bytes_to_file(self, filename: str, content: bytes) -> None:
        try:
            with open(filename, "wb") as f:
                f.write(content)
            os.chmod(filename, 0o755)
        except Exception as err:
            self.logger.exception(
                f"An error was raised in CertificateService while writing to file {filename}, error: {err}"
            )

    def _get_self_sign_issuer_name(self) -> x509.Name:
        return x509.Name(
            [
                x509.NameAttribute(NameOID.COUNTRY_NAME, ISSUER_COUNTRY_NAME),
                x509.NameAttribute(NameOID.ORGANIZATION_NAME, ISSUER_ORGANIZATION_NAME),
            ]
        )

    def generate_certificate(
        self,
    ) -> str:
        # create dir for private key and certificate
        os.makedirs(self._cert_path, mode=0o777, exist_ok=True)

        # generate key pair
        key_pair = self._crypt_gateway.generate_key_pair(
            self._cert_request.key_algorithm,
            self._cert_request.key_size,
            self._cert_request.passphrase,
        )

        # write private key to file

        self._write_bytes_to_file(self._private_key_file, key_pair.private_key_pem)

        # generate certificate
        subject_name = map_certificaterequest_to_x509name(self._cert_request)
        subject_alternative_name = map_certificaterequest_to_x509subjectalternativename(
            self._cert_request
        )
        certificate_pem = self._crypt_gateway.generate_certificate_pem(
            subject_name=subject_name,
            issuer_name=self._get_self_sign_issuer_name(),
            key_pair_details=key_pair,
            passphrase=self._cert_request.passphrase,
            days_valid=self._cert_request.days_valid or DEFAULT_DAYS_VALID,
            subject_alternative_name=subject_alternative_name,
        )

        # write certificate to file

        self._write_bytes_to_file(self._certificate_file, certificate_pem)
        self.logger.info(f"Write private key and certificate to {self._cert_path}")
        return self._cert_path



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import logging
import re
from typing import Any, Dict, List, Optional

from fbpcp.entity.container_instance import ContainerInstance
from fbpcp.error.pcp import InvalidParameterError, PcpError
from fbpcp.gateway.ecs import ECSGateway
from fbpcp.metrics.emitter import MetricsEmitter
from fbpcp.service.container import ContainerService
from fbpcp.util.aws import split_container_definition

AWS_API_INPUT_SIZE_LIMIT = 100  # AWS API Call Capacity Limit


class AWSContainerService(ContainerService):
    def __init__(
        self,
        region: str,
        cluster: str,
        subnets: Optional[List[str]] = None,
        access_key_id: Optional[str] = None,
        access_key_data: Optional[str] = None,
        config: Optional[Dict[str, Any]] = None,
        metrics: Optional[MetricsEmitter] = None,
        session_token: Optional[str] = None,
    ) -> None:
        self.logger: logging.Logger = logging.getLogger(__name__)
        self.region = region
        self.cluster = cluster
        self.subnets = subnets
        self.ecs_gateway = ECSGateway(
            region, access_key_id, access_key_data, config, metrics, session_token
        )

    def get_region(
        self,
    ) -> str:
        return self.region

    def get_cluster(
        self,
    ) -> str:
        return self.cluster

    def create_instance(
        self,
        container_definition: str,
        cmd: str,
        env_vars: Optional[Dict[str, str]] = None,
    ) -> ContainerInstance:
        task_definition, container = split_container_definition(container_definition)

        if not self.subnets:
            raise PcpError(
                "No subnets specified. It's required to create container instances."
            )

        return self.ecs_gateway.run_task(
            task_definition, container, cmd, self.cluster, self.subnets, env_vars
        )

    def create_instances(
        self,
        container_definition: str,
        cmds: List[str],
        env_vars: Optional[Dict[str, str]] = None,
    ) -> List[ContainerInstance]:
        instances = [
            self.create_instance(container_definition, cmd, env_vars) for cmd in cmds
        ]
        self.logger.info(
            f"AWSContainerService created {len(instances)} containers successfully"
        )
        return instances

    def get_instance(self, instance_id: str) -> Optional[ContainerInstance]:
        return self.ecs_gateway.describe_task(self.cluster, instance_id)

    def get_instances(
        self, instance_ids: List[str]
    ) -> List[Optional[ContainerInstance]]:
        """Get one or more container instances

        Args:
            instance_ids: a list of the container instances.

        Returns:
            A list of Optional, in the same order as the input ids. For example, if
            users pass 3 instance_ids and the second instance could not be found,
            then returned list should also have 3 elements, with the 2nd elements being None.
        """
        id_batches = [
            instance_ids[i : i + AWS_API_INPUT_SIZE_LIMIT]
            for i in range(0, len(instance_ids), AWS_API_INPUT_SIZE_LIMIT)
        ]
        container_batches = [
            self.ecs_gateway.describe_tasks(self.cluster, ids) for ids in id_batches
        ]
        return sum(container_batches, [])

    def cancel_instance(self, instance_id: str) -> None:
        return self.ecs_gateway.stop_task(cluster=self.cluster, task_id=instance_id)

    def cancel_instances(self, instance_ids: List[str]) -> List[Optional[PcpError]]:
        res = []
        for instance_id in instance_ids:
            try:
                res.append(self.cancel_instance(instance_id))
            except PcpError as err:
                res.append(err)

        return res

    def get_current_instances_count(self) -> int:
        cluster = self.ecs_gateway.describe_cluster(self.cluster)
        return cluster.running_tasks + cluster.pending_tasks

    def validate_container_definition(self, container_definition: str) -> None:
        if not re.fullmatch(r"[\w\-/:]+?:\d+#[\w\-/]+?", container_definition):
            raise InvalidParameterError(
                "Parameter container_definition must be in format <task definition name>:<revision>#<container name>"
            )



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from enum import Enum


class ChecksumType(Enum):
    MD5 = "MD5"
    SHA256 = "SHA256"
    BLAKE2B = "BLAKE2B"



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import logging
import os
import uuid
from pathlib import Path
from typing import Optional

from fbpcp.service.storage import StorageService
from onedocker.common.core_dump_handler import CoreDumpHandler

CORE_DUMP_FILE_PREFIX = "core"


class AWSCoreDumpHandler(CoreDumpHandler):
    def __init__(self, storage_svc: StorageService) -> None:
        self.storage_svc = storage_svc
        self.logger: logging.Logger = logging.getLogger("__name__")

    def locate_core_dump_file(self) -> Optional[str]:
        """Find the absolute path of the generated core dump file
        AWS: /proc/sys/kernel/core_pattern: core
        The core dump file will be generated in the current
        working dir. File name starts with 'core'.
        """
        cwd = os.getcwd()
        file_list = [f for f in os.listdir(cwd) if f.startswith(CORE_DUMP_FILE_PREFIX)]

        if len(file_list) == 0:
            self.logger.error(f"Core dump file not found in {cwd}!")
            return None

        # One core file is generated for each run
        file_path = os.path.join(cwd, file_list[0])

        if not Path(file_path).is_file:
            # If the file path points to a non-file type
            # return None
            self.logger.error(f"{file_path} is not a file.")
            return None

        self.logger.info(f"Core dump file locates in {file_path}")
        return file_path

    def upload_core_dump_file(self, core_dump_file_path: str, upload_dest: str) -> None:
        """Upload the core dump file to S3
        Valid upload_dest for AWS is the S3 bucket e.g. http://test.s3.us-west-2.amazonaws.com/
        """
        # uploaded core file named as core.uuid
        file_suffix = str(uuid.uuid4())
        s3_dest = f"{upload_dest}core.{file_suffix}"
        self.storage_svc.copy(core_dump_file_path, s3_dest)
        self.logger.info(f"Core dump file was uploaded to {s3_dest}")



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import json
import logging
from typing import Dict, List, Union

from fbpcp.service.storage import StorageService
from onedocker.entity.checksum_type import ChecksumType
from onedocker.service.checksum import LocalChecksumGenerator

# Default checksum types
DEFAULT_CHECKSUM_TYPES: List[ChecksumType] = [
    ChecksumType.MD5,
    ChecksumType.SHA256,
    ChecksumType.BLAKE2B,
]

# Package Info Dict Tags
PACKAGE_NAME = "Package Name"
PACKAGE_VERSION = "Package Version"
PACKAGE_CHECKSUMS = "Checksums"


class AttestationService:
    def __init__(self, storage_svc: StorageService, repository_path: str) -> None:
        self.logger: logging.Logger = logging.getLogger(__name__)
        self.checksum_generator = LocalChecksumGenerator()
        self.storage_svc = storage_svc
        self.repository_path = repository_path

    def _build_attestation_repository_path(
        self,
        package_name: str,
        version: str,
    ) -> str:
        return f"{self.repository_path}{package_name}/{version}.json"

    def _format_package_info(
        self,
        package_name: str,
        version: str,
        checksums: Dict[str, str],
    ) -> Dict[str, Union[str, Dict[str, str]]]:
        package_info = {}
        package_info[PACKAGE_NAME] = package_name
        package_info[PACKAGE_VERSION] = version
        package_info[PACKAGE_CHECKSUMS] = checksums
        return package_info

    def _upload_checksum(
        self,
        package_name: str,
        version: str,
        checksums: Dict[str, str],
    ) -> None:
        # Put together package into a json format
        package_info = self._format_package_info(package_name, version, checksums)

        # Construct file information - (name and contents)
        file_path = self._build_attestation_repository_path(package_name, version)
        file_contents = json.dumps(package_info, indent=4)

        # upload contents to set repo path
        self.storage_svc.write(file_path, file_contents)

    def track_binary(
        self,
        binary_path: str,
        package_name: str,
        version: str,
    ) -> None:
        """
        This Function generates then uploads checksums for passed in local binary

        Args:
            binary_path:    Local file path pointing to the package
            package_name:   Package Name to use when uploading file to checksum repository
            version:        Package Version to relay while uploading file to checksum repository
        """
        # Generates checksums
        self.logger.info(f"Generating checksums for binary at {binary_path}")
        checksums: Dict[str, str] = self.checksum_generator.generate_checksums(
            binary_path=binary_path,
            checksum_algorithms=DEFAULT_CHECKSUM_TYPES,
        )

        # Upload checksums and package info to set repo path
        self.logger.info(f"Uploading checksums for package {package_name}: {version}")
        self._upload_checksum(
            package_name=package_name,
            version=version,
            checksums=checksums,
        )

    def verify_binary(
        self,
        binary_path: str,
        package_name: str,
        version: str,
        checksum_algorithm: ChecksumType,
    ) -> None:
        """
        This functions generates a checksum for a local file and then compares the generated value to what is stored in the checksum repository for the given package_name and version

        Args:
            binary_path:            Local file path pointing to the package
            package_name:           Package Name to use when downlading the checksum file from checksum repository
            version:                Package Version to relay while downloading the checksum file from checksum repository
            checksum_algorithm:     Checksum algorithm that should be used while attesting local binary
        """
        # Build file path
        file_path = self._build_attestation_repository_path(package_name, version)

        if not self.storage_svc.file_exists(file_path):
            self.logger.info(
                f"Untracked package {package_name}: {version}. Skipping Attestion."
            )
            return None
        # Retrieve file info and parse contents
        file_contents = self.storage_svc.read(file_path)
        package_info = json.loads(file_contents)

        # Verify that file contents are for desired package
        self.logger.info("Attesting correct package information was retrived")
        if package_info[PACKAGE_NAME] != package_name:
            raise ValueError(
                f"Package Name {package_info[PACKAGE_NAME]} in file is different than passed in Name {package_name}"
            )
        if package_info[PACKAGE_VERSION] != version:
            raise ValueError(
                f"Package Version {package_info[PACKAGE_VERSION]} in file is different than passed in Version {version}"
            )

        # Process downloaded file and generate a local checksum
        checksums: Dict[str, str] = self.checksum_generator.generate_checksums(
            binary_path=binary_path,
            checksum_algorithms=[checksum_algorithm],
        )
        package_checksums = package_info[PACKAGE_CHECKSUMS]

        # Verify Checksum Details
        self.logger.info("Attesting binary integrity")
        if checksum_algorithm.name not in package_checksums:
            raise ValueError(
                f"Package Checksum Algorithms {package_checksums.keys()} does not contain passed in Checksum Algorithm {checksum_algorithm.name}"
            )
        if (
            checksums[checksum_algorithm.name]
            != package_checksums[checksum_algorithm.name]
        ):
            raise ValueError(
                f"Package Checksum {package_checksums[checksum_algorithm.name]} for Checksum Algorithm {checksum_algorithm.name} does not match actual binary Checksum {checksums[checksum_algorithm.name]}"
            )
        self.logger.info("Binary successfully attested")
